/*
** ʰץȥåܡ
*/
#include <cmath>
#include "Trackball.h"

#ifndef M_PI
#  define M_PI 3.14159265358979323846
#endif

#define SCALE (2.0 * M_PI) // ޥа֢žѤδ

/*
** r <- p x q
*/
static void qmul(double r[], const double p[], const double q[])
{
  r[0] = p[0] * q[0] - p[1] * q[1] - p[2] * q[2] - p[3] * q[3];
  r[1] = p[0] * q[1] + p[1] * q[0] + p[2] * q[3] - p[3] * q[2];
  r[2] = p[0] * q[2] - p[1] * q[3] + p[2] * q[0] + p[3] * q[1];
  r[3] = p[0] * q[3] + p[1] * q[2] - p[2] * q[1] + p[3] * q[0];
}

/*
** žѴ r <- ˥ q
*/
static void qrot(double r[], double q[])
{
  double x2 = q[1] * q[1] * 2.0;
  double y2 = q[2] * q[2] * 2.0;
  double z2 = q[3] * q[3] * 2.0;
  double xy = q[1] * q[2] * 2.0;
  double yz = q[2] * q[3] * 2.0;
  double zx = q[3] * q[1] * 2.0;
  double xw = q[1] * q[0] * 2.0;
  double yw = q[2] * q[0] * 2.0;
  double zw = q[3] * q[0] * 2.0;
  
  r[ 0] = 1.0 - y2 - z2;
  r[ 1] = xy + zw;
  r[ 2] = zx - yw;
  r[ 4] = xy - zw;
  r[ 5] = 1.0 - z2 - x2;
  r[ 6] = yz + xw;
  r[ 8] = zx + yw;
  r[ 9] = yz - xw;
  r[10] = 1.0 - x2 - y2;
  r[ 3] = r[ 7] = r[11] = r[12] = r[13] = r[14] = 0.0;
  r[15] = 1.0;
}

/*
** ȥåܡν
** ץνΤȤǼ¹Ԥ
*/
Trackball::Trackball(void)
{
  /* ɥåǤϤʤ */
  drag = false;

  /* ñ̥˥ */
  cq[0] = 1.0;
  cq[1] = 0.0;
  cq[2] = 0.0;
  cq[3] = 0.0;
  
  /* žν */
  rt[ 0] = 1.0; rt[ 1] = 0.0; rt[ 2] = 0.0; rt[ 3] = 0.0;
  rt[ 4] = 0.0; rt[ 5] = 1.0; rt[ 6] = 0.0; rt[ 7] = 0.0;
  rt[ 8] = 0.0; rt[ 9] = 0.0; rt[10] = 1.0; rt[11] = 0.0;
  rt[12] = 0.0; rt[13] = 0.0; rt[14] = 0.0; rt[15] = 1.0;
}

/*
** ȥåܡ뤹ΰ
** Reshape Хå (resize) Ǽ¹Ԥ
*/
void Trackball::region(int w, int h)
{
  /* ޥݥ󥿰֤ΥɥŪ֤ؤδ */
  sx = 1.0 / (double)w;
  sy = 1.0 / (double)h;
}

/*
** ɥå
** ޥܥ򲡤Ȥ˼¹Ԥ
*/
void Trackball::start(int x, int y)
{
  /* ɥå */
  drag = true;

  /* ɥåϿ */
  cx = x;
  cy = y;
}

/*
** ɥå
** ޥΥɥå˼¹Ԥ
*/
void Trackball::motion(int x, int y)
{
  if (drag) {
    double dx, dy, a;
    
    /* ޥݥ󥿤ΰ֤Υɥåϰ֤Ѱ */
    dx = (x - cx) * sx;
    dy = (y - cy) * sy;
    
    /* ޥݥ󥿤ΰ֤Υɥåϰ֤εΥ */
    a = sqrt(dx * dx + dy * dy);
    
    if (a != 0.0) {
      double ar = a * SCALE * 0.5;
      double as = sin(ar) / a;
      double dq[4] = { cos(ar), dy * as, dx * as, 0.0 };
      
      /* ˥ݤƲž */
      qmul(tq, dq, cq);
      
      /* ˥󤫤žѴ */
      qrot(rt, tq);
    }
  }
}

/*
** 
** ޥܥΥȤ˼¹Ԥ
*/
void Trackball::stop(int x, int y)
{
  /* ɥåλˤž */
  motion(x, y);

  /* ž¸ */
  cq[0] = tq[0];
  cq[1] = tq[1];
  cq[2] = tq[2];
  cq[3] = tq[3];

  /* ɥåλ */
  drag = false;
}

/*
** žѴ᤹
** ͤ glMultMatrixd() ʤɤǻѤƥ֥Ȥž
*/
const double *Trackball::rotation(void)
{
  return rt;
}
