#include <iostream>
#include <fstream>
#include <cstdlib>
#include <cmath>

#if defined(WIN32)
//#  pragma comment(linker, "/subsystem:\"windows\" /entry:\"mainCRTStartup\"")
#  include "glut.h"
#  include "glext.h"
PFNGLLOADTRANSPOSEMATRIXDPROC glLoadTransposeMatrixd;
#elif defined(__APPLE__) || defined(MACOSX)
#  include <GLUT/glut.h>
#else
#  define GL_GLEXT_PROTOTYPES
#  include <GL/glut.h>
#endif

/*
** VF[_IuWFNg
*/
#include "glsl.h"
static GLuint shader0, shader1;
static GLint texture, cubemap;

/*
**gbN{[
*/
#include "Trackball.h"
static Trackball *tb1, *tb2;
static int btn;

/*
** 
*/
#include "Box.h"
static Box *box;

/*
** OBJ t@C
*/
#include "Obj.h"
static Obj *obj;

/*
** eNX`
*/
#define TEXWIDTH  128                           /* eNX`̕@@@ */
#define TEXHEIGHT 128                           /* eNX`̍@@ */
static GLuint texname[2];                       /* eNX`iԍj */
static const char *texfile[] = {                /* eNX`t@C */
  "room2ny.raw", /*  */
  "room2nz.raw", /*  */
  "room2px.raw", /* E */
  "room2pz.raw", /* O */
  "room2nx.raw", /*  */
  "room2py.raw", /*  */
};
static const int target[] = {                /* eNX`̃^[Qbg */
  GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
  GL_TEXTURE_CUBE_MAP_NEGATIVE_Z,
  GL_TEXTURE_CUBE_MAP_POSITIVE_X,
  GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
  GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
  GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
};

/*
** VF[_vO̍쐬
*/
static GLuint loadShader(const char *vert, const char *frag)
{
  /* VF[_IuWFNg̍쐬 */
  GLuint vertShader = glCreateShader(GL_VERTEX_SHADER);
  GLuint fragShader = glCreateShader(GL_FRAGMENT_SHADER);
  
  /* VF[_̃\[XvO̓ǂݍ */
  if (readShaderSource(vertShader, vert)) exit(1);
  if (readShaderSource(fragShader, frag)) exit(1);
  
  /* VF[_vÕRpC^Nʂ𓾂ϐ */
  GLint compiled, linked;

  /* o[ebNXVF[_̃\[XvÕRpC */
  glCompileShader(vertShader);
  glGetShaderiv(vertShader, GL_COMPILE_STATUS, &compiled);
  printShaderInfoLog(vertShader);
  if (compiled == GL_FALSE) {
    std::cerr << "Compile error in vertex shader." << std::endl;
    exit(1);
  }
  
  /* tOgVF[_̃\[XvÕRpC */
  glCompileShader(fragShader);
  glGetShaderiv(fragShader, GL_COMPILE_STATUS, &compiled);
  printShaderInfoLog(fragShader);
  if (compiled == GL_FALSE) {
    std::cerr << "Compile error in fragment shader." << std::endl;
    exit(1);
  }
  
  /* vOIuWFNg̍쐬 */
  GLuint gl2Program = glCreateProgram();
  
  /* VF[_IuWFNg̃VF[_vOւ̓o^ */
  glAttachShader(gl2Program, vertShader);
  glAttachShader(gl2Program, fragShader);
  
  /* VF[_IuWFNg̍폜 */
  glDeleteShader(vertShader);
  glDeleteShader(fragShader);
  
  /* VF[_vÕN */
  glLinkProgram(gl2Program);
  glGetProgramiv(gl2Program, GL_LINK_STATUS, &linked);
  printProgramInfoLog(gl2Program);
  if (linked == GL_FALSE) {
    std::cerr << "Link error" << std::endl;
    exit(1);
  }

  return gl2Program;
}

/*
** 
*/
static void init(void)
{
  /* ̃IuWFNg𐶐 */
  box = new Box(500.0f, 500.0f, 500.0f);

  /* OBJ t@C̓ǂݍ */
  obj = new Obj("bunny.obj");

  /* gbN{[pIuWFNg̐ */
  tb1 = new Trackball;
  tb2 = new Trackball;

  /* eNX`Q */
  glGenTextures(2, texname);
  
  /* eNX`摜̓[hPʂɋlߍ܂Ă */
  glPixelStorei(GL_UNPACK_ALIGNMENT, 4);

  /* Ỗ̗eNX`̊蓖āiWj */
  glBindTexture(GL_TEXTURE_2D, texname[0]);
  glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA, TEXWIDTH * 8, TEXHEIGHT, 0,
    GL_RGBA, GL_UNSIGNED_BYTE, 0);
  
  /* eNX`摜̓ǂݍ */
  for (int i = 0; i < 6; ++i) {
    std::ifstream file;

    file.open(texfile[i], std::ios::binary);
    if (file) {
      GLubyte image[TEXHEIGHT * TEXWIDTH * 4]; // eNX`摜̓ǂݍݗp

      file.read(reinterpret_cast<char *>(image), sizeof image);
      file.close();

      /* Ỗ̗eNX`̒u */
      glTexSubImage2D(GL_TEXTURE_2D, 0, TEXWIDTH * i, 0, TEXWIDTH, TEXHEIGHT,
        GL_RGBA, GL_UNSIGNED_BYTE, image);

      /* L[u}bsÕeNX`̊蓖 */
      glBindTexture(GL_TEXTURE_CUBE_MAP, texname[1]);
      glTexImage2D(target[i], 0, GL_RGBA, TEXWIDTH, TEXHEIGHT, 0, 
        GL_RGBA, GL_UNSIGNED_BYTE, image);

      /* ݒΏۂỖ̗eNX`ɖ߂ */
      glBindTexture(GL_TEXTURE_2D, texname[0]);
    }
  }
  
  /* eNX`gEk@̎w */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  
  /* eNX`̌JԂ@̎w */
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);

  /* ݒΏۂL[u}bsÕeNX`ɐ؂ւ */
  glBindTexture(GL_TEXTURE_CUBE_MAP, texname[1]);

  /* eNX`gEk@̎w */
  glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
  glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
  
  /* eNX`̌JԂ@̎w */
  glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
  glTexParameteri(GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
  
  /* ݒΏۂ𖳖eNX`ɖ߂ */
  glBindTexture(GL_TEXTURE_2D, 0);
  
  /* GLSL ̏ */
  if (glslInit()) exit(1);

  /* VF[_vO̍쐬 */
  shader0 = loadShader("replace.vert", "replace.frag");
  shader1 = loadShader("reflect.vert", "reflect.frag");
  
  /* uniform ϐ̈ʒuo */
  texture = glGetUniformLocation(shader0, "texture");
  cubemap = glGetUniformLocation(shader1, "cubemap");

  /* ݒ */
  glClearColor(0.3, 0.3, 1.0, 0.0);
  glEnable(GL_DEPTH_TEST);
  glDisable(GL_CULL_FACE);
  
#if defined(WIN32)
  glLoadTransposeMatrixd =
    (PFNGLLOADTRANSPOSEMATRIXDPROC)wglGetProcAddress("glLoadTransposeMatrixd");
#endif
}

/*
** V[̕`
*/
static void scene(void)
{
  /* ݒΏۂỖ̗eNX`ɐ؂ւ */
  glBindTexture(GL_TEXTURE_2D, texname[0]);
  
  /* ̃eNX`̃VF[_vOKp */
  glUseProgram(shader0);

  /* eNX`jbgOw肷 */
  glUniform1i(texture, 0);

  /* ` */
  glPushMatrix();
  glMultMatrixd(tb2->rotation());
  box->draw();
  glPopMatrix();

  /* ݒΏۂL[u}bsÕeNX`ɐ؂ւ*/
  glBindTexture(GL_TEXTURE_CUBE_MAP, texname[1]);

  /* L[u}bsÕVF[_vOKp */
  glUseProgram(shader1);

  /* eNX`jbgOw肷 */
  glUniform1i(cubemap, 0);

  /* eNX`ϊsɃgbN{[̉] */
  glMatrixMode(GL_TEXTURE);
  glLoadTransposeMatrixd(tb2->rotation());
  glMatrixMode(GL_MODELVIEW);

  /* _菭ɃIuWFNg`ăgbN{[̉] */
  glPushMatrix();
  glTranslated(0.0, 0.0, -200.0);
  glMultMatrixd(tb1->rotation());
  obj->draw();
  glPopMatrix();
  
  /* eNX`ϊsɖ߂ */
  glMatrixMode(GL_TEXTURE);
  glLoadIdentity();
  glMatrixMode(GL_MODELVIEW);

  /* ݒΏۂ𖳖eNX`ɖ߂ */
  glBindTexture(GL_TEXTURE_2D, 0);
}


/****************************
** GLUT ̃R[obN֐ **
****************************/

static void display(void)
{
  /* ʃNA */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  /* fr[ϊs̐ݒ */
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  
  /* V[̕` */
  scene();
  
  /* _uobt@O */
  glutSwapBuffers();
}

static void resize(int w, int h)
{
  /* gbN{[͈ */
  tb1->region(w, h);
  tb2->region(w, h);
  
  /* EBhEŜr[|[gɂ */
  glViewport(0, 0, w, h);
  
  /* ϊs̎w */
  glMatrixMode(GL_PROJECTION);
  
  /* ϊs̏ */
  glLoadIdentity();
  gluPerspective(60.0, (double)w / (double)h, 100.0, 500.0);
}

static void idle(void)
{
  /* ʂ̕`ւ */
  glutPostRedisplay();
}

static void mouse(int button, int state, int x, int y)
{
  btn = button;

  switch (btn) {
  case GLUT_LEFT_BUTTON:
    if (state == GLUT_DOWN) {
      /* gbN{[Jn */
      tb1->start(x, y);
      glutIdleFunc(idle);
    }
    else {
      /* gbN{[~ */
      tb1->stop(x, y);
      glutIdleFunc(0);
    }
    break;
  case GLUT_RIGHT_BUTTON:
    if (state == GLUT_DOWN) {
      /* gbN{[Jn */
      tb2->start(x, y);
      glutIdleFunc(idle);
    }
    else {
      /* gbN{[~ */
      tb2->stop(x, y);
      glutIdleFunc(0);
    }
    break;
  default:
    break;
  }
}

static void motion(int x, int y)
{
  switch (btn) {
  case GLUT_LEFT_BUTTON:
    /* gbN{[ړ */
    tb1->motion(x, y);
    break;
  case GLUT_RIGHT_BUTTON:
    /* gbN{[ړ */
    tb2->motion(x, y);
    break;
  default:
    break;
  }
}

static void keyboard(unsigned char key, int x, int y)
{
  switch (key) {
  case 'q':
  case 'Q':
  case '\033':
    /* ESC  q  Q ^CvI */
    exit(0);
  default:
    break;
  }
}

/*
** CvO
*/
int main(int argc, char *argv[])
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_RGBA | GLUT_DEPTH | GLUT_DOUBLE);
  glutCreateWindow(argv[0]);
  glutDisplayFunc(display);
  glutReshapeFunc(resize);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(keyboard);
  init();
  glutMainLoop();
  return 0;
}
